"""Module of constants used to establish the connection to the data."""

from pathlib import Path

DATA_REPO_URL = "https://raw.githubusercontent.com/sosia-dev/sosia-data/"
_cache_folder = Path.home() / ".cache" / "sosia/"
FIELD_SOURCE_MAP = _cache_folder / "field_sources_list.csv"
SOURCE_INFO = _cache_folder / "source_info.csv"
DEFAULT_DATABASE = _cache_folder / 'main.sqlite'
DEFAULT_LOG = _cache_folder / 'sosia.log'

DB_TABLES = {
    "author_citations":
        {"columns": (("auth_id", "INTEGER"), ("year", "INTEGER"),
                     ("n_cits", "INTEGER")),
         "primary": ("auth_id", "year")},
    "author_data":
        {"columns": (("auth_id", "INTEGER"), ("year", "INTEGER"),
                     ("first_year", "INTEGER"), ("n_pubs", "INTEGER"),
                     ("n_coauth", "INTEGER")),
         "primary": ("auth_id", "year")},
    "author_info":
        {"columns": (("auth_id", "INTEGER"), ("surname", "TEXT"),
                     ("givenname", "TEXT"), ("documents", "INTEGER"),
                     ("areas", "TEXT")),
         "primary": ("auth_id",)},
    "sources":
        {"columns": (("source_id", "INTEGER"), ("year", "INTEGER"),
                     ("auids", "TEXT")),
         "primary": ("source_id", "year")}}



"""Tests for processing.querying module."""

from string import Template

from sosia.processing import base_query, count_citations, create_queries,\
    query_pubs_by_sourceyear, stacked_query

test_id = 53164702100
year = 2017


def test_base_query():
    q = f"AU-ID({test_id}) AND PUBYEAR BEF {year}"
    size = base_query("docs", q, size_only=True)
    assert size == 5


def test_base_query_author():
    query = f"AU-ID({test_id})"
    size = base_query("author", query, size_only=True)
    assert size == 1


def test_count_citations():
    identifier = ["55208373700", "55208373700"]
    count1 = count_citations(identifier, 2017)
    assert count1 == 24
    eids = ["2-s2.0-84959420483", "2-s2.0-84949113230"]
    count2 = count_citations(eids, 2017, exclusion_ids=identifier)
    assert count2 == 1
    eids_long = eids * 100
    count3 = count_citations(eids_long, 2017, exclusion_ids=identifier)
    assert count3 == 1


def test_create_queries_long():
    # Set variables
    group = list(range(1, 2000))
    template = Template(f"SOURCE-ID($fill) AND PUBYEAR IS {year+1}")
    joiner = " OR "
    maxlen = 200
    # Run test
    received = create_queries(group, joiner, template, maxlen)
    query_maxlen = max([len(q[0]) for q in received])
    # Compare
    assert isinstance(received, list)
    assert isinstance(received[0], tuple)
    assert query_maxlen <= maxlen
    expected = 'SOURCE-ID(1 OR 10 OR 100 OR 1000 OR 1001 OR 1002 OR 1003 OR '\
               '1004 OR 1005 OR 1006 OR 1007 OR 1008 OR 1009 OR 101 OR 1010 '\
               'OR 1011 OR 1012 OR 1013 OR 1014 OR 1015 OR 1016 OR 1017) '\
               'AND PUBYEAR IS 2018'
    sub_group = ['1', '10', '100', '1000', '1001', '1002', '1003', '1004',
                 '1005', '1006', '1007', '1008', '1009', '101', '1010', '1011',
                 '1012', '1013', '1014', '1015', '1016', '1017']
    assert received[0][0] == expected
    assert received[0][1] == sub_group


def test_create_queries_short():
    # Set variables
    group = list(range(1, 2000))
    template = Template(f"SOURCE-ID($fill) AND PUBYEAR IS {year+1}")
    joiner = " OR "
    maxlen = 1
    # Run test
    received = create_queries(group, joiner, template, maxlen)
    group_maxlen = max([len(q[1]) for q in received])
    # Compare
    assert isinstance(received, list)
    assert isinstance(received[0], tuple)
    assert group_maxlen <= maxlen
    expected = 'SOURCE-ID(1) AND PUBYEAR IS 2018'
    sub_group = ['1']
    assert received[0][0] == expected
    assert received[0][1] == sub_group


def test_query_sources_by_year(refresh_interval):
    # Test a journal and year
    res = query_pubs_by_sourceyear([22900], 2010, refresh=refresh_interval)
    assert res["source_id"].unique() == [22900]
    assert res["year"].unique() == [2010]
    assert isinstance(res["auids"][0], str)
    assert len(res["auids"][0]) > 0
    # Test a journal and year that are not in Scopus
    res = query_pubs_by_sourceyear([22900], 1969, refresh=refresh_interval)
    assert res.empty
    # Test a large query (>5000 results)
    source_ids = [13703, 13847, 13945, 14131, 14150, 14156, 14204, 14207,
                  14209, 14346, 14438, 14536, 14539, 15034, 15448, 15510, 15754]
    res = query_pubs_by_sourceyear(source_ids, 1984, refresh=refresh_interval)
    assert 1 < res.dropna(subset=["auids"]).shape[0] <= len(source_ids)
    assert res.columns.tolist() == ['source_id', 'year', 'auids']
    assert isinstance(res["auids"][0], str)
    assert len(res["auids"][0]) > 0


def test_query_sources_by_year_stacked(refresh_interval):
    # Test a journal and year
    res = query_pubs_by_sourceyear([22900], 2010, refresh=refresh_interval,
                                   stacked=True)
    assert res["source_id"].unique() == [22900]
    assert res["year"].unique() == [2010]
    assert isinstance(res["auids"][0], str)
    assert len(res["auids"][0]) > 0
    # Test a journal and year that are not in Scopus
    res = query_pubs_by_sourceyear([22900], 1969, refresh=refresh_interval,
                                   stacked=True)
    assert res.empty
    # Test a large query (>5000 results)
    source_ids = [13703, 13847, 13945, 14131, 14150, 14156, 14204, 14207,
                  14209, 14346, 14438, 14536, 14539, 15034, 15448, 15510, 15754]
    res = query_pubs_by_sourceyear(source_ids, 1984, refresh=refresh_interval, stacked=True)
    assert 1 < res.dropna(subset=["auids"]).shape[0] <= len(source_ids)
    assert res.columns.tolist() == ['source_id', 'year', 'auids']
    assert isinstance(res["auids"][0], str)
    assert len(res["auids"][0]) > 0


def test_stacked_query():
    group = [18400156716, 19300157101, 19400157208, 19400157312, 19500157223,
             19600166213, 19700175482, 19700182353, 19800188009, 19900193211]
    template = Template(f"SOURCE-ID($fill) AND PUBYEAR IS {year+1}")
    res = stacked_query(group, template, joiner=" OR ", q_type="docs",
                        refresh=False, stacked=True, verbose=False)
    assert len(res) == 791



"""Tests for processing.utils module."""

import pandas as pd

from sosia.processing import chunk_list, compute_margins, compute_overlap, \
    flat_set_from_df


def test_chunk_list():
    result1 = chunk_list(range(1999, 2007 + 1), 3)
    expected1 = [[1999, 2000, 2001], [2002, 2003, 2004], [2005, 2006, 2007]]
    assert result1 == expected1
    result2 = chunk_list(range(1999, 2006 + 1), 3)
    expected2 = [[1999, 2000, 2001], [2002, 2003, 2004], [2005, 2006]]
    assert result2 == expected2
    result3 = chunk_list(range(1999, 2005 + 1), 3)
    expected3 = [[1999, 2000, 2001], [2002, 2003, 2004, 2005]]
    assert result3 == expected3


def test_compute_margins():
    assert compute_margins(5, 1) == (4, 6)
    assert compute_margins(10, 0.09) == (9, 11)
    assert compute_margins(150, 200) == (0, 350)


def test_compute_overlap():
    set1 = set("abc")
    set2 = set("cde")
    assert compute_overlap(set1, set2) == 1


def test_flat_set_from_df():
    d = {'col1': [[1, 2], [10, 20]], "col2": ["a", "b"]}
    df = pd.DataFrame(d)
    expected = [1, 2, 10, 20]
    received = sorted(list(flat_set_from_df(df, "col1")))
    assert received == expected



import sys
import os
from datetime import date

cwd = os.getcwd()
project_root = os.path.dirname(cwd)
sys.path.insert(0, project_root)

# -- Project information -----------------------------------------------------

project = "sosia"
author = "Michael E. Rose and Stefano H. Baruffaldi"
copyright = f"2017-{date.today().year} {author}"

import sosia

version = sosia.__version__
release = sosia.__version__
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    'sphinx_copybutton'
]
copybutton_prompt_text = ">>> "
autodoc_member_order = "groupwise"
templates_path = ["_templates"]
source_suffix = ".rst"
master_doc = "index"
language = 'en'
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
pygments_style = "sphinx"


# -- Options for HTML output -------------------------------------------------

html_theme = "alabaster"
html_theme_options = {
    "github_user": "sosia-dev",
    "github_repo": "sosia",
    "github_banner": "true",
    "github_button": "true",
    "github_type": "star",
}
html_static_path = ["_static"]
html_sidebars = {"**": ["about.html", "navigation.html", "searchbox.html"]}


# -- Options for HTMLHelp output ---------------------------------------------

html_show_sourcelink = True
autoclass_content = "both"

htmlhelp_basename = "sosiadoc"


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

latex_documents = [
    (
        master_doc,
        "sosia.tex",
        "sosia Documentation",
        "Michael E. Rose and Stefano H. Baruffaldi",
        "manual",
    )
]


# -- Options for manual page output ------------------------------------------

man_pages = [(master_doc, "sosia", "sosia Documentation", [author], 1)]


# -- Options for Texinfo output ----------------------------------------------

texinfo_documents = [
    (
        master_doc,
        "sosia",
        "sosia Documentation",
        author,
        "sosia",
        "One line description of project.",
        "Miscellaneous",
    )
]


# -- Options for Epub output -------------------------------------------------

epub_title = project

epub_exclude_files = ["search.html"]


# -- Extension configuration -------------------------------------------------

# -- Options for intersphinx extension ---------------------------------------

intersphinx_mapping = {"https://docs.python.org/": None}



"""Module with functions for retrieving and processing author data."""

from string import Template

import pandas as pd
from tqdm import tqdm

from sosia.processing.extracting import extract_yearly_author_data
from sosia.processing.caching import insert_data, retrieve_from_author_table, \
    retrieve_authors_from_sourceyear
from sosia.processing.querying import count_citations, stacked_query, \
    query_pubs_by_sourceyear
from sosia.utils import custom_print


def get_author_info(authors, conn, verbose=False, refresh=False) -> pd.DataFrame:
    """Get author information from author_info table and add missing
    information via Author Search API.

    Parameters
    ----------
    authors : list
       List of Scopus Author IDs to search.

    conn : sqlite3 connection
        Standing connection to a SQLite3 database.

    verbose : bool (optional, default=False)
        Whether to print information on the search progress.

    refresh : bool, int (optional, default=False)
        How to handle existing information in database and on disk.  If True,
        or int is passed, will replace all matching information in database
        with information on disk.  If int is passed, information on disk
        will be refreshed if older than int days.  If True, will refresh
        information on disk in any case.

    Returns
    -------
    info : DataFrame
        DataFrame with general information on requested authors.
    """
    # Retrieve information from SQL database
    authors = pd.DataFrame(authors, columns=["auth_id"], dtype="uint64")
    info, missing = retrieve_from_author_table(authors, conn,
        table="author_info", refresh=refresh)

    # Query missing records and insert at the same time
    if missing:
        params = {"group": missing, "refresh": refresh, "joiner": ") OR AU-ID(",
                  "q_type": "author", "template": Template("AU-ID($fill)"),
                  "stacked": True, "verbose": verbose}
        text = f"Downloading information for {len(missing):,} candidates..."
        custom_print(text, verbose)
        res = stacked_query(**params)
        if not res:  # Likely author IDs do not exist anymore
            return info
        res = pd.DataFrame(res)
        res = res.drop_duplicates(subset="eid")
        res["auth_id"] = res['eid'].str.split('-').str[-1].astype("int64")
        res = res[info.columns]
        insert_data(res, conn, table="author_info")
        info = pd.concat([info, res])
    return info


def get_author_data(group, conn, verbose=False, refresh=False) -> pd.DataFrame:
    """Get author information from author_data table and add missing
    information via Scopus Search API.

    Parameters
    ----------
    group : list of str
        Scopus IDs of authors to be filtered.

    conn : sqlite3 connection
        Standing connection to a SQLite3 database.

    verbose : bool (optional, default=False)
        Whether to print information on the search progress.

    refresh : bool, int (optional, default=False)
        How to handle existing information in database and on disk.  If True,
        or int is passed, will replace all matching information in database
        with information on disk.  If int is passed, information on disk
        will be refreshed if older than int days.  If True, will refresh
        information on disk in any case.

    Returns
    -------
    data : DataFrame
        DataFrame with yearly information on requested authors.
    """
    # Retrieve information from SQL database
    authors = pd.DataFrame({"auth_id": group})
    data, missing = retrieve_from_author_table(authors, conn,
        table="author_data", refresh=refresh)

    # Add to database
    if missing:
        text = f"Querying Scopus for information for {len(missing):,} " \
               "authors..."
        custom_print(text, verbose)
        to_add = []
        for auth_id in tqdm(missing, disable=not verbose):
            new = extract_yearly_author_data(auth_id, refresh=refresh)
            to_add.append(new)
        to_add = pd.concat(to_add)
        insert_data(to_add, conn, table="author_data")
        data = pd.concat([data, to_add])
    return data


def get_authors_from_sourceyear(df, conn, refresh=False, *args, **kwargs):
    """Get authors publishing in specified sourced in specified years.

    Handles retrieving data, and in case of missing data querying for it
    and inserting it into the SQL database.

    Parameters
    ----------
    df : DataFrame
        DataFrame of source-year-combinations to be searched for.

    conn : sqlite3 connection
        Standing connection to an SQLite3 database.

    refresh : bool, int (optional, default=False)
        How to handle existing information in database and on disk.  If True,
        or int is passed, will replace all matching information in database
        with information on disk.  If int is passed, information on disk
        will be refreshed if older than int days.  If True, will refresh
        information on disk in any case.

    *args, **kwargs : tuple or dict (optional)
        Additional options passed on to `query_pubs_by_sourceyear()`:
        `verbose`, and `stacked`.

    Returns
    -------
    data : DataFrame
        DataFrame in format ("source_id", "year", "auids"), where
        auids is a string of author IDs joined on semicolon.
    """
    # Retrieve information from SQL database
    drop = refresh is not False  # becomes True for all values unless False
    data, missing = retrieve_authors_from_sourceyear(df, conn, drop=drop)

    # Download and add missing data
    to_add = pd.DataFrame()
    empty = []
    for year in missing["year"].unique():
        subset = missing[missing["year"] == year]
        sources = subset["source_id"].unique()
        new = query_pubs_by_sourceyear(sources, year, refresh=refresh,
                                       *args, **kwargs)
        no_info = set(sources) - set(new["source_id"].unique())
        assert new["source_id"].nunique() + len(no_info) == len(sources)
        empty.extend([(s, year) for s in no_info])
        to_add = pd.concat([to_add, new])

    # Insert new information and information on missing data
    if empty:
        sources, years = list(zip(*empty))
        d = {"source_id": sources, "year": years, "auids": [""] * len(sources)}
        to_add = pd.concat([to_add, pd.DataFrame(d)])
    if not to_add.empty:
        insert_data(to_add, conn, table="sources")

    # Return data
    data = pd.concat([data, to_add])
    data = data[data["auids"] != ""]
    data["auids"] = data["auids"].str.split(";")
    return data


def get_citations(authors, year, conn, verbose=False, refresh=False):
    """Get citations net of self-citations in particular `year` for
    a group of `authors`.

    Parameters
    ----------
    authors : list
       List of Scopus Author IDs for whom to return citation information.

    year : int
        The year in which citations must be counted.

    conn : sqlite3 connection
        Standing connection to a SQLite3 database.

    verbose : bool (optional, default=False)
        Whether to print information on the search progress.

    refresh : bool, int (optional, default=False)
        Whether to refresh cached results (if they exist) or not, with
        Scopus data that is at most `refresh` days old (True = 0).

    Returns
    -------
    citations : DataFrame
        Data on the provided authors.
    """
    df = pd.DataFrame({"auth_id": authors, "year": year})
    citations, missing = retrieve_from_author_table(df, conn,
        table="author_citations", refresh=refresh)
    # Add citations
    if missing:
        text = f"Counting citations of {len(missing):,} candidates..."
        custom_print(text, verbose)
        to_add = []
        for auth_id in tqdm(missing, disable=not verbose):
            n_cits = count_citations([auth_id], year + 1)
            to_add.append(n_cits)
        to_add = pd.DataFrame({"auth_id": missing, "year": year, "n_cits": to_add})
        insert_data(to_add, conn, table="author_citations")
        citations = pd.concat([citations, to_add])
    # Filter on year
    citations = citations[citations["year"] == year].drop(columns="year")
    return citations.astype("uint64")



"""Module with functions for querying and processing data from Scopus."""

from string import Template

import pandas as pd
from pybliometrics.scopus.exception import Scopus400Error
from tqdm import tqdm

from sosia.establishing import ScopusLogger
from sosia.processing.constants import AUTHOR_SEARCH_MAX_COUNT, QUERY_MAX_LEN, \
    RESEARCH_TYPES
from sosia.utils import custom_print


def base_query(q_type, query, refresh=False, view="COMPLETE", fields=None,
               size_only=False):
    """Wrapper function to perform a particular search query.

    Parameters
    ----------
    q_type : str
        Determines the query search that will be used.  Allowed values:
        "author", "docs".

    query : str
        The query string.

    refresh : bool or int (optional, default=False)
            Whether to refresh cached results (if they exist) or not. If
            int is passed, results will be refreshed if they are older
            than that value in number of days.

    fields : list of field names (optional, default=None)
        Fields in the Scopus query that must always present.  To be passed
        onto pybliometrics.scopus.ScopusSearch.  Will be ignored
        when q_type = "author".

    size_only : bool (optional, default=False)
        Whether to not download results and return the number
        of results instead.

    Returns
    -------
    res : list of namedtuples (if size_only is False) or int
        Documents represented by namedtuples as returned from scopus or
        number of search results.

    Raises
    ------
    ValueError:
        If q_type is none of the allowed values.
    """

    from pybliometrics.scopus import AuthorSearch, ScopusSearch, init

    init()

    params = {"query": query, "refresh": refresh, "download": not size_only}

    if q_type == "author":
        params["count"] = AUTHOR_SEARCH_MAX_COUNT
        with ScopusLogger("Author Search", params) as sl:
            au = AuthorSearch(**params)
            sl.scopus_obj = au
        if size_only:
            return au.get_results_size()
        else:
            return au.authors or []
    elif q_type == "docs":
        params["integrity_fields"] = fields
        params["view"] = view
        if size_only:
            with ScopusLogger("Scopus Search", params) as sl:
                ss = ScopusSearch(**params)
                sl.scopus_obj = ss
            return ss.get_results_size()
        try:
            with ScopusLogger("Scopus Search", params) as sl:
                ss = ScopusSearch(**params)
                sl.scopus_obj = ss
            docs = ss.results or []
            docs = [d for d in docs if d.subtype in RESEARCH_TYPES]
            return docs
        except AttributeError:
            params.pop("integrity_fields")
            params["refresh"] = True
            with ScopusLogger("Scopus Search", params) as sl:
                ss = ScopusSearch(**params)
                sl.scopus_obj = ss
            docs = ss.results or []
            return docs


def count_citations(search_ids, pubyear, exclusion_ids=None):
    """Auxiliary function to count non-self citations up to a year."""
    if not exclusion_ids:
        exclusion_ids = search_ids
    exclusion_ids = [str(i) for i in exclusion_ids]
    q = f"REF({' OR '.join([str(i) for i in search_ids])}) AND PUBYEAR BEF "\
        f"{pubyear} AND NOT AU-ID({') AND NOT AU-ID('.join(exclusion_ids)})"
    # Download if too long
    if len(q) > QUERY_MAX_LEN:
        template = Template(f"REF($fill) AND PUBYEAR BEF {pubyear} AND NOT"
                            f" (AU-ID({') OR AU-ID('.join(exclusion_ids)}))")
        queries = create_queries(search_ids, " OR ", template, QUERY_MAX_LEN)
        res = []
        for q in queries:
            r = long_query(q, "docs", template, view="STANDARD")
            res.extend(r)
        res = pd.DataFrame(res)
        return res["eid"].nunique()
    return base_query("docs", q, size_only=True)


def create_queries(group, joiner, template, maxlen):
    """Create queries combining `maxlen` entities of a `group` to search for.

    Parameters
    ----------
    group : list
        Entities (authors, documents, etc.) to search for on Scopus.

    joiner : str
        On which the group elements should be joined to fill the query.

    template : string.Template()
        A string template with one parameter named `fill` which will be used
        as search query.

    maxlen : int
        The maximum length a query can be. If equal 1, one element at the time
        is used per query.

    Returns
    -------
    queries : list of tuples
        A list of tuples where the first element of each tuple is a query
        and the second is the list of elements searched by the query.
    """
    group = sorted([str(g) for g in group])
    queries = []
    start = 0
    for i, g in enumerate(group):
        sub_group = group[start:i+2]
        query = template.substitute(fill=joiner.join(sub_group))
        if maxlen == 1 or len(query) > maxlen or i+1 == len(group):
            sub_group = group[start:i+1]
            query = template.substitute(fill=joiner.join(sub_group))
            queries.append((query, sub_group))
            start = i + 1
    return queries


def long_query(query, q_type, template, *args, **kwargs):
    """Run one query from create_queries output, and revert to
    one-by-one queries of each element if Scopus400Error is returned.

    Parameters
    ----------
    query : str
        The query string.

    q_type : str
        Determines the query search that will be used.  Allowed values:
        "author", "docs".

    template : string.Template()
        A string template with one parameter named `fill` which will be used
        as search query.

    *args, **kwargs : tuple or dict (optional)
        Additional parameters to be passed to `base_query()`: `refresh`,
        and `view`.
    """
    try:
        return base_query(q_type, query[0], *args, **kwargs)
    except Scopus400Error:
        res = []
        for element in query[1]:
            q = template.substitute(fill=element)
            res = base_query(q_type, q, *args, **kwargs)
            res.extend(res)
        return res


def query_pubs_by_sourceyear(source_ids, year, verbose=False, *args, **kwargs):
    """Get authors lists for each source in a particular year.

    Parameters
    ----------
    source_ids : list
        List of Scopus IDs of sources to search.

    year : str or int
        The year of the search.

    verbose : bool or int (default = False)
        Whether to report on the progress of the process.

    *args, **kwargs : tuple or dict (optional)
        Additional options passed to stacked_query(): `stacked`, and
        `refresh`.
    """
    # Search authors
    msg = f"... parsing Scopus information for {year}..."
    custom_print(msg, verbose)
    q = Template(f"SOURCE-ID($fill) AND PUBYEAR IS {year}")
    res = stacked_query(
        group=[str(x) for x in sorted(source_ids)],
        joiner=" OR ",
        q_type="docs",
        verbose=verbose,
        template=q,
        *args, **kwargs
    )

    # Verify data is not empty
    dummy = pd.DataFrame(columns=["source_id", "year", "auids"])
    if res:
        res = pd.DataFrame(res).dropna(subset=["author_ids"])
        if res.empty:
            return dummy
    else:
        return dummy

    # Group data
    data = {"source_id": [],
            "year": year,
            "auids": []}
    for source_id, subset in res.groupby("source_id"):
        groups = subset["author_ids"].dropna().tolist()
        authors = sorted(set([item for sublist in groups for item in sublist.split(";")]))
        data["source_id"].append(int(source_id))
        data["auids"].append(";".join(authors))
    data = pd.DataFrame(data)
    return data


def stacked_query(group, template, joiner, q_type, stacked=False,
                  verbose=False, *args, **kwargs):
    """Auxiliary function to query list of items.

    Parameters
    ----------
    group : list of str
        Scopus IDs (of authors or sources) for which the stacked query should
        be conducted.

    template : string.Template()
        A string template with one parameter named `fill` which will be used
        as search query.

    joiner : str
        On which the group elements should be joined to fill the query.

    q_type : str
        Determines the query search that will be used.  Allowed values:
        "author", "docs".

    stacked: bool (optional, default=False)
        If True search for queries as close as possible to the maximum length
        QUERY_MAX_LEN.  If False, search elements in group one by one.

    verbose : bool (optional, default=False)
        If True, prints a progress bar for the download sections.

    *args, **kwargs : dict (optional)
        Additional options passed on to `long_query()`: `refresh`.

    Returns
    -------
    res : list
        A list of namedtuples representing publications.
    """
    maxlen = 1
    if stacked:
        maxlen = QUERY_MAX_LEN
    queries = create_queries(group, joiner, template, maxlen)
    res = []
    for q in tqdm(queries, disable=not verbose):
        res.extend(long_query(q, q_type, template, *args, **kwargs))
    return res




"""This module is the entry point for the `sosia` package."""

from importlib.metadata import version

__version__ = version("sosia")

__citation__ = 'Rose, Michael E. and Stefano H. Baruffaldi: "Finding '\
    'Doppelgängers in Scopus: How to Build Scientists Control Groups Using '\
    'Sosia", Max Planck Institute for Innovation & Competition Research '\
    'Paper No. 20-20.'

from sosia.classes import Original
from sosia.establishing import *



"""This module imports the necessary functions and classes for establishing connections
and accessing data in `sosia`.
"""
from sosia.establishing.constants import *
from sosia.establishing.database import *
from sosia.establishing.fields_sources import *
from sosia.establishing.logger import *



"""Module that provides functions for downloading and storing field-source information from the
sosia-dev/sosia-data repository.
"""

import pandas as pd

from sosia.establishing.constants import DATA_REPO_URL, FIELD_SOURCE_MAP, \
    SOURCE_INFO
from sosia.utils import custom_print


def get_field_source_information(verbose: bool = False) -> None:
    """Download two files from sosia-dev/sosia-data repository:
    1. List of Scopus source IDs with additional information
    2. Mapping of sources to ASJC codes

    Parameters
    ----------
    verbose : bool (optional, default=False)
        Whether to report on the progress of the process.
    """
    fname = DATA_REPO_URL + "main/sources/source_info.csv"
    info = pd.read_csv(fname, index_col=0)
    SOURCE_INFO.parent.mkdir(parents=True, exist_ok=True)
    info.to_csv(SOURCE_INFO)

    fname = DATA_REPO_URL + "main/sources/field_sources_map.csv"
    fields = pd.read_csv(fname, index_col=0)
    fields.to_csv(FIELD_SOURCE_MAP)

    text = f"Stored information for {info.shape[0]:,} sources as well as "\
        f"{fields.shape[0]:,} field-source assignments in {SOURCE_INFO.parent}"
    custom_print(text, verbose)



"""Tests for processing.caching.retrieving module."""

from itertools import product

import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal
from pybliometrics.scopus import AuthorSearch

from sosia.establishing import connect_database, make_database
from sosia.processing import retrieve_from_author_table, \
    retrieve_authors_from_sourceyear, query_pubs_by_sourceyear
from sosia.processing.caching import insert_data


def test_retrieve_from_author_table(test_cache):
    make_database(test_cache, drop=True)
    conn = connect_database(test_cache, verbose=False)
    # Variables
    expected_auth = [53164702100, 57197093438]
    df = pd.DataFrame(expected_auth, columns=["auth_id"], dtype="int64")
    expected_cols = ['auth_id', 'surname', 'givenname', 'documents', 'areas']
    # Retrieve data
    incache, missing = retrieve_from_author_table(df, conn, table="author_info")
    assert incache.shape[0] == 0
    assert incache.columns.to_list() == expected_cols
    assert missing == expected_auth


def test_retrieve_from_author_table_insert(test_cache, refresh_interval):
    make_database(test_cache, drop=True)
    conn = connect_database(test_cache, verbose=False)
    # Variables
    expected_auth = [53164702100, 57197093438]
    search_auth = [55317901900]
    expected_cols = ['auth_id', 'eid', 'surname', 'initials', 'givenname',
                     'affiliation', 'documents', 'affiliation_id', 'city',
                     'country', 'areas']
    # Insert data
    q = f"AU-ID({') OR AU-ID('.join([str(a) for a in expected_auth])})"
    res = pd.DataFrame(AuthorSearch(q, refresh=refresh_interval).authors)
    res["auth_id"] = res["eid"].str.split("-").str[-1].astype("int64")
    res["affiliation_id"] = res["affiliation_id"].astype(float)
    res = res[expected_cols]
    insert_data(res, conn, table="author_info")
    # Retrieve data
    df = pd.DataFrame(expected_auth + search_auth, columns=["auth_id"],
                      dtype="uint64")
    incache, missing = retrieve_from_author_table(df, conn, table="author_info")
    assert incache.shape[0] == 2
    assert missing == [55317901900]


def test_retrieve_authors_from_sourceyear(test_cache, refresh_interval):
    make_database(test_cache, drop=True)
    conn = connect_database(test_cache, verbose=False)
    # Variables
    expected_sources = [22900]
    expected_years = [2005, 2010]
    df = pd.DataFrame(product(expected_sources, expected_years),
                      columns=["source_id", "year"], dtype="int64")
    # Populate cache
    expected = query_pubs_by_sourceyear(expected_sources, expected_years[0],
                                        refresh=refresh_interval)
    expected["source_id"] = expected["source_id"].astype(np.int64)
    expected = expected.sort_values("auids").reset_index(drop=True)
    expected = expected[['source_id', 'year', 'auids']]
    insert_data(expected, conn, table="sources")
    # Retrieve from cache
    incache, missing = retrieve_authors_from_sourceyear(df, conn)
    incache = incache.sort_values("auids").reset_index(drop=True)
    assert_frame_equal(incache, expected)
    assert_frame_equal(missing, df.tail(1).reset_index(drop=True))



"""Tests for processing.initializing module."""

from pandas import DataFrame

from sosia.classes import Scientist
from sosia.processing.initializing import add_source_names, \
    read_fields_sources_list


def test_add_source_names(test_cache):
    s = Scientist([55208373700], 2017, db_path=test_cache)
    expected = [(14351, "Brain Research Reviews"),
                (18632, "Progress in Brain Research")]
    ids, names = zip(*expected)
    received = add_source_names(ids, s.source_names)
    assert received == expected


def test_read_fields_sources_list():
    sources, names = read_fields_sources_list()
    assert isinstance(sources, DataFrame)
    assert isinstance(sources, DataFrame)



"""Module with constants used in sosia for processing."""

ASJC_2D = {10: "MULT", 11: "AGRI", 12: "ARTS", 13: "BIOC", 14: "BUSI",
           15: "CENG", 16: "CHEM", 17: "COMP", 18: "DECI", 19: "EART",
           20: "ECON", 21: "ENER", 22: "ENGI", 23: "ENVI", 24: "IMMU",
           25: "MATE", 26: "MATH", 27: "MEDI", 28: "NEUR", 29: "NURS",
           30: "PHAR", 31: "PHYS", 32: "PSYC", 33: "SOCI", 34: "VETE",
           35: "DENT", 36: "HEAL"}

QUERY_MAX_LEN = 2000
AUTHOR_SEARCH_MAX_COUNT = 200

RESEARCH_TYPES = {"ar", "bk", "ch", "cp", "cr", "no", "re", "sh"}



"""Tests for class `Original`."""

import warnings
from collections import namedtuple

import pandas as pd

warnings.filterwarnings("ignore")
FIRST_YEAR_MARGIN = 1

# Expected matches
fields = "ID name first_year num_coauthors num_publications num_citations "\
         "affiliation_country language num_cited_refs"
Match = namedtuple("Match", fields)
MATCHES = [
    Match(
        ID=55567912500,
        name='Eling, Katrin',
        first_year=2013,
        num_coauthors=9,
        num_publications=8,
        num_citations=56,
        affiliation_country='Netherlands',
        language='eng',
        num_cited_refs=0)
]


def test_search_sources(original1):
    original1.define_search_sources()
    search_sources = original1.search_sources
    assert len(search_sources) == 206
    assert (20206, "Academy of Management Review") in search_sources
    assert (15143, "Regional Studies") in search_sources


def test_search_sources_change(original1):
    backup = original1.search_sources
    expected = [(14351, "Brain Research Reviews"),
                (18632, "Progress in Brain Research")]
    original1.search_sources, _ = zip(*expected)
    assert original1.search_sources == expected
    original1.search_sources = backup


def test_candidates(original1, refresh_interval):
    original1.identify_candidates_from_sources(frequency=2, refresh=refresh_interval,
                                               first_year_margin=FIRST_YEAR_MARGIN)
    recieved = original1.candidates
    assert isinstance(recieved, list)
    assert 700 <= len(recieved) <= 800


def test_candidates_stacked(original1, refresh_interval):
    original1.identify_candidates_from_sources(frequency=2, stacked=True,
                                               refresh=refresh_interval,
                                               first_year_margin=FIRST_YEAR_MARGIN)
    recieved = original1.candidates
    assert isinstance(recieved, list)
    assert 700 <= len(recieved) <= 800


def test_filter_candidates(original1, refresh_interval):
    original1.filter_candidates(refresh=refresh_interval, cits_margin=0.15,
                                same_discipline=True, pub_margin=0.2,
                                first_year_margin=FIRST_YEAR_MARGIN,
                                coauth_margin=0.2)
    expected = [m.ID for m in MATCHES]
    assert original1.matches == expected


def test_inform_matches(original1, refresh_interval):
    fields = ["first_year", "num_coauthors", "num_publications",
              "num_citations", "affiliation_country", "language",
              "num_cited_refs"]
    original1.inform_matches(refresh=refresh_interval, fields=fields)
    recieved = original1.matches
    assert len(recieved) == len(MATCHES)
    assert isinstance(recieved, list)
    df_r = pd.DataFrame(recieved)
    df_r["num_cited_refs"] = df_r["num_cited_refs"].round(3)
    df_m = pd.DataFrame(MATCHES)
    df_m["num_cited_refs"] = df_m["num_cited_refs"].round(3)
    pd.testing.assert_frame_equal(df_r, df_m)




"""Module with helper functions."""


def custom_print(text, verbose):
    """Print text if verbose."""
    if verbose:
        print(text)


def get_ending(num, plural="s"):
    """Return singular or plural ending."""
    if num == 1:
        return ""
    else:
        return plural


def validate_param(value, label, accepted_types=(int, float)):
    """Validates that a given value is of the specified types."""
    if value is not None and not isinstance(value, accepted_types):
        accepted_types_names = ", ".join(t.__name__ for t in accepted_types)
        msg = f"Argument '{label}' must be one of the following " \
              f"types: {accepted_types_names}"
        raise TypeError(msg)



"""Main module of `sosia` containing the `Original` class."""

from math import ceil
from itertools import product
from pathlib import Path
from typing import Iterable, Literal, Optional, Union

from pandas import DataFrame
from typing_extensions import Self

from sosia.classes import Scientist
from sosia.establishing import create_logger, DEFAULT_LOG
from sosia.processing import add_source_names, chunk_list, compute_margins, \
    flat_set_from_df, get_author_data, get_author_info, \
    get_authors_from_sourceyear, get_citations, generate_filter_message, \
    inform_matches
from sosia.utils import accepts, custom_print, get_ending, validate_param


class Original(Scientist):
    """Representation of a scientist for whom to find a control scientist."""
    @property
    def candidates(self) -> Optional[list[int]]:
        """The set of authors that might be matches to the scientist.  The
        set contains the intersection of all authors publishing in the
        treatment year as well as authors publishing around the year of first
        publication.  Some authors with too many publications in the
        treatment year and authors having published too early are removed.

        Notes
        -----
        Property is initiated via .identify_candidates_from_sources().
        """
        return self._candidates

    @property
    def matches(self) -> Optional[list]:
        """List of Scopus IDs or list of namedtuples representing matches
        of the original scientist in the treatment year.

        Notes
        -----
        Property is initiated via .find_matches().
        """
        return self._matches

    @property
    def search_sources(self) -> Optional[Union[set, list, tuple]]:
        """The set of sources comparable to those of the Original.

        A source (journal, book, etc.) is comparable if it belongs to
        the Original's main field, and if the types of the sources are those
        the Original publishes in.

        Notes
        -----
        Property is initiated via .define_search_sources().
        """
        return self._search_sources

    @search_sources.setter
    @accepts((set, list, tuple))
    def search_sources(self, val: Union[set, list, tuple]) -> None:
        self._search_sources = add_source_names(val, self.source_names)

    def __init__(
        self,
        scientist: Union[str, int, list[Union[str, int]]],
        match_year: Union[str, int],
        eids: Optional[list[Union[str, int]]] = None,
        refresh: Union[bool, int] = False,
        db_path: Optional[Union[str, Path]] = None,
        log_path: Optional[Union[str, Path]] = None,
        verbose: Optional[bool] = False
    ) -> None:
        """Representation of a scientist for whom to find matches (= Original).

        Parameters
        ----------
        scientist : str, int or list of str or int
            Scopus Author ID, or list of Scopus Author IDs, of the scientist
            to find a control scientist for.

        match_year : str or numeric
            Year in which the comparison takes place.  Control scientist will
            be matched on trends and characteristics of the original
            scientist up to this year.

        eids : list (optional, default=None)
            A list of Scopus EIDs of the publications to consinder.  If it
            is provided, all properties will be derived from them, and the
            control group is based on them. If None, will use all
            research-type publications published until the match year.

        refresh : boolean (optional, default=False)
            Whether to refresh cached results (if they exist) or not.  If int
            is passed, results will be refreshed if they are older than
            that value in number of days.

        db_path : str or pathlib.Path (optional, default=None)
            The path of the SQLite database to connect to.  If None,
            will default to `~/.cache/sosia/main.sqlite`.  Will be created
            if the database doesn't exist.

        log_path : str or pathlib.Path (optional, default=None)
            The path of the log file using logging.  If None, will default
             to `~/.cache/sosia/sosia.log`.

        verbose : bool (optional, default=False)
            Whether to report on the initialization process.
        """
        # Variables
        if not isinstance(scientist, list):
            scientist = [scientist]
        self.identifier = [int(auth_id) for auth_id in scientist]
        self.match_year = int(match_year)
        self.eids = eids
        self.sql_fname = db_path
        self._search_sources = None
        self._candidates = None
        self._matches = None

        # Create logger
        log_path = log_path or DEFAULT_LOG
        create_logger(Path(log_path))

        # Instantiate superclass to load private variables
        Scientist.__init__(self, self.identifier, match_year, refresh=refresh,
                           db_path=self.sql_fname, verbose=verbose)

    def identify_candidates_from_sources(
        self,
        first_year_margin: int,
        frequency: Optional[int] = None,
        stacked: bool = False,
        verbose: bool = False,
        refresh: Union[bool, int] = False,
    ) -> Self:
        """Define a search group of authors based on their publication
        activity in the search sources between the first year and the
        match year.

        Parameters
        ----------
        first_year_margin : int
            The left margin for year of first publication to identify
            match candidates.

        frequency : int (optional, default=None)
            The maximum gap in number of years between publications of
            suitable candidates, i.e. the average frequency with which they
            publish in these sources.  If not given, will take the average
            frequency (rounded up) of the Orignal:
            `max[1, (match_year - first_year) / number of publications.]`
            Must not be smaller than the first_year_margin.
            To find candidates, the method considers chunks of consecutive
            volumes, and requires candidates to publish at least once in
            each (!) of these chunks.  That is, it generates sets of authors
            publishing in the search sources during specific years, and
            considers only the intersection of these. If the last chunk is
            smaller than half the target chunk size, it will be merged
            with the previous chunk.

        stacked : bool (optional, default=False)
            Whether to combine searches in few queries or not.  Cached
            files will most likely not be reusable.  Set to True if you
            query in distinct fields or to minimize API key usage.

        verbose : bool (optional, default=False)
            Whether to report on the progress of the process.

        refresh : bool (optional, default=False)
            refresh : bool or int (optional, default=False)
            Whether to refresh cached results (if they exist) or not.  If
            int is passed, results will be refreshed if they are older
            than that value in number of days.

        Notes
        -----
        If candidates have been identified, they are accessible through
        property .candidates.
        """
        # Checks
        if not self.search_sources:
            text = "No search sources defined.  Please run "\
                   ".define_search_sources() first."
            raise RuntimeError(text)
        validate_param(first_year_margin, "first_year_margin", (int,))
        if not frequency:
            n_pubs = len(self.publications)
            frequency = max(1, ceil((self.match_year - self.first_year) / n_pubs))

        # Define variables
        search_sources, _ = zip(*self.search_sources)
        text = f"Identifying candidates using up to {len(search_sources):,} sources..."
        custom_print(text, verbose)

        # Get years to look through
        years = range(self.first_year, self.match_year)
        chunks = chunk_list(years, frequency)
        chunks[0] = range(self.first_year - first_year_margin,
                          max(chunks[0]) + 1)

        # Get authors
        groups = []
        for years in chunks:
            volumes = DataFrame(product(search_sources, years),
                                columns=["source_id", "year"])
            authors = get_authors_from_sourceyear(
                volumes,
                self.sql_conn,
                stacked=stacked,
                refresh=refresh,
                verbose=verbose
            )
            groups.append(flat_set_from_df(authors, "auids"))

        # Compile group
        candidates = set.intersection(*groups)
        candidates = set(map(int, candidates))
        candidates -= set(self.identifier)
        candidates -= set(self.coauthors)

        # Finalize
        self._candidates = sorted(candidates)
        n_candidates = len(candidates)
        text = f"Found {n_candidates:,} candidate{get_ending(n_candidates)}"
        custom_print(text, verbose)
        return self

    def define_search_sources(
            self,
            verbose: bool = False,
            mode: Literal["narrow", "wide"] = "narrow"
    ) -> Self:
        """Define search sources related to the Original.

        Search sources are the set of sources where sosia will search for
        possible candidates.  Search source are of the same types (journal,
        conference proceeding, etc.) the Original published in, and is
        related to the main field (ASCJ-4) of the Original.

        Parameters
        ----------
        verbose : bool (optional, default=False)
            Whether to report on the progress of the process.

        mode : str (optional, default="narrow")
            Accepted values: "narrow", "wide".
            A "narrow" definition of search sources excludes search sources
            that are also associated to fields (ASJC-4) not among the fields
            of the Original.  A "wide" defintion includes all those sources.

        Notes
        -----
        Search sources are available through property `.search_sources`.
        """
        if mode not in {"narrow", "wide"}:
            raise TypeError("Argument mode must be 'narrow' or 'wide'.")

        field_df = self.field_source
        info_df = self.source_info
        # Sources in scientist's main field
        mask_same_field = field_df["asjc"] == self.main_field[0]
        same_field = set(field_df[mask_same_field]["source_id"].unique())
        # Types of Scientist's sources
        own_source_ids, _ = zip(*self.sources)
        same_sources = info_df["source_id"].isin(own_source_ids)
        main_types = info_df[same_sources]["type"].unique()
        mask_same_types = info_df["type"].isin(main_types)
        same_type = info_df[mask_same_types]["source_id"].unique()
        # Select source IDs
        selected_ids = same_field.intersection(same_type)
        selected = field_df[field_df["source_id"].isin(selected_ids)].copy()
        grouped = selected.groupby("source_id")["asjc"].unique().to_frame()
        # Deselect sources with alien fields
        if mode == "narrow":
            fields = set(self.fields)
            no_alien_field = grouped["asjc"].apply(lambda s: len(set(s) - fields) == 0)
            grouped = grouped[no_alien_field]
        grouped = grouped.drop(columns="asjc")
        # Add source names
        sources = grouped.join(info_df.set_index("source_id")["title"])
        sources = set(sources.reset_index().itertuples(index=False, name=None))
        # Add own sources
        sources.update(self.sources)
        # Finalize
        self._search_sources = sorted(sources, key=lambda x: x[0])
        text = f"Found {len(sources):,} sources of " \
               f"type{get_ending(len(main_types))} " \
               f"{', '.join(main_types)} matching main field "\
               f"{self.main_field[0]} {mode}ly"
        custom_print(text, verbose)
        return self

    def filter_candidates(
            self,
            first_year_margin: Optional[int] = None,
            pub_margin: Optional[Union[float, int]] = None,
            coauth_margin: Optional[Union[float, int]] = None,
            cits_margin: Optional[Union[float, int]] = None,
            same_discipline: Optional[bool] = False,
            verbose: bool = False,
            refresh: Union[bool, int] = False
    ) -> None:
        """Find matches within candidates based on up to five criteria:
        1. Work mainly in the same discipline (as of date of retrieval)
        2. Started publishing in about the same year
        3. Have about the same number of publications in the match year
        4. Have about the same number of coauthors in the match year
        5. Have about the same number of citations in the match year

        Parameters
        ----------
        first_year_margin : numeric (optional, default=None)
            The left and right margin for year of first publication to match
            candidates and the scientist on. If the value is not given,
            sosia will not filter on the first year of publication.

        pub_margin : numeric (optional, default=None)
            The left and right margin for the number of publications to match
            candidates and the scientist on.  If the value is a float,
            it is interpreted as percentage of the scientist's number of
            publications and the resulting value is rounded up.  If the value
            is an integer, it is interpreted as fixed number of publications.
            If the value is not given, sosia will not filter on the number
            of publications.

        coauth_margin : numeric (optional, default=None)
            The left and right margin for the number of coauthors to match
            candidates and the scientist on.  If the value is a float,
            it is interpreted as percentage of the scientists number of
            coauthors and the resulting value is rounded up.  If the value
            is an integer, it is interpreted as fixed number of coauthors.
            If the value is not given, sosia will not filter on the number
            of coauthors.

        cits_margin : numeric (optional, default=None)
            The left and right margin for the number of citations to match
            candidates and the scientist on.  If the value is a float,
            it is interpreted as percentage of the scientists number of
            publications and the resulting value is rounded up.  If the value
            is an integer, it is interpreted as fixed number of citations.
            If the value is not given, sosia will not filter on the number
            of citations.

        same_discipline : boolean (optional, default=False)
            Whether to restrict candidates to the same main discipline (ASJC2)
            as the original scientist or not.

        verbose : bool (optional, default=False)
            Whether to report on the progress of the process.

        refresh : bool or int (optional, default=False)
            Whether to refresh cached results (if they exist) or not. If
            int is passed, results will be refreshed if they are older
            than that value in number of days.

        Notes
        -----
        Matches are available through property `.matches`.
        """
        # Checks
        if not self.candidates:
            text = "No candidates defined.  Please define candidates first."
            raise RuntimeError(text)
        validate_param(first_year_margin, "first_year_margin", (int,))
        validate_param(pub_margin, "pub_margin")
        validate_param(coauth_margin, "coauth_margin")
        validate_param(cits_margin, "cits_margin")

        # Find matches
        group = self.candidates
        text = f"Filtering {len(group):,} candidates..."
        custom_print(text, verbose)

        # First round of filtering: minimum publications and main field
        if same_discipline or pub_margin is not None:
            info = get_author_info(group, self.sql_conn, verbose=verbose,
                                   refresh=refresh)
            if same_discipline:
                same_discipline = info['areas'].str.startswith(self.main_field[1])
                info = info[same_discipline]
                text = (f"... left with {info.shape[0]:,} candidates with same "
                        f"main discipline ({self.main_field[1]})")
                custom_print(text, verbose)
            if pub_margin is not None:
                min_papers = compute_margins(len(self.publications), pub_margin)[0]
                enough_pubs = info['documents'].astype(int) >= min_papers
                info = info[enough_pubs]
                text = (f"... left with {info.shape[0]:,} candidates with "
                        f"sufficient total publications ({min_papers:,})")
                custom_print(text, verbose)
            group = sorted(info["auth_id"].unique())

        # Second round of filtering: first year, publication count, coauthor count
        second_round = (
                (first_year_margin is not None) or
                (pub_margin is not None) or
                (coauth_margin is not None)
        )
        if second_round:
            data = get_author_data(group=group, verbose=verbose,
                                   conn=self.sql_conn, refresh=refresh)
            data = data[data["year"] <= self.match_year]
            if first_year_margin is not None:
                _years = compute_margins(self.first_year, first_year_margin)
                similar_start = data["first_year"].between(*_years)
                data = data[similar_start]
                text = generate_filter_message(data['auth_id'].nunique(), _years,
                                               "year of first publication")
                custom_print(text, verbose)
            data = (data.drop(columns="first_year")
                        .drop_duplicates("auth_id", keep="last"))
            if pub_margin is not None:
                _npapers = compute_margins(len(self.publications), pub_margin)
                similar_pubcount = data["n_pubs"].between(*_npapers)
                data = data[similar_pubcount]
                text = generate_filter_message(data.shape[0], _npapers,
                                               "number of publications")
                custom_print(text, verbose)
            if coauth_margin is not None:
                _ncoauth = compute_margins(len(self.coauthors), coauth_margin)
                similar_coauthcount = data["n_coauth"].between(*_ncoauth)
                data = data[similar_coauthcount]
                text = generate_filter_message(data.shape[0], _ncoauth,
                                               "number of coauthors")
                custom_print(text, verbose)
            group = sorted(data["auth_id"].unique())

        # Third round of filtering: citations
        if cits_margin is not None:
            citations = get_citations(group, self.year, refresh=refresh,
                                      verbose=verbose, conn=self.sql_conn)
            _ncits = compute_margins(self.citations, cits_margin)
            similar_citcount = citations["n_cits"].between(*_ncits)
            citations = citations[similar_citcount]
            text = generate_filter_message(citations.shape[0], _ncits,
                                           "number of citations")
            custom_print(text, verbose)
            group = sorted(citations['auth_id'].unique())

        # Status update
        n_matches = len(group)
        text = f"Found {n_matches:,} match{get_ending(n_matches, 'es')}"
        custom_print(text, verbose)
        self._matches = sorted([int(auth_id) for auth_id in group])

    def inform_matches(
        self,
        fields: Optional[Iterable] = None,
        verbose: bool = False,
        refresh: Union[bool, int] = False,
    ) -> None:
        """Add information to matches to aid in selection process.

        Parameters
        ----------
        fields : iterable (optional, default=None)
            Which information to provide.  Allowed values are "first_name",
            "surname", "first_year", "last_year", "num_coauthors",
            "num_publications", "num_citations", "subjects",
            "affiliation_country", "affiliation_id", "affiliation_name",
            "affiliation_type", "language", "num_cited_refs".  If None, will
            use all available fields.

        verbose : bool (optional, default=False)
            Whether to report on the progress of the process.

        refresh : bool or int (optional, default=False)
            Whether to refresh cached results (if they exist) or not. If
            int is passed, results will be refreshed if they are older
            than that value in number of days.

        Notes
        -----
        Matches including corresponding information are available through
        property `.matches`.

        Raises
        ------
        ValueError
            If `fields` contains invalid keywords.
        """
        # Checks
        if not self._matches:
            text = "No matches defined.  Please run .find_matches() first."
            raise RuntimeError(text)
        allowed_fields = ["first_name", "surname", "first_year", "last_year",
                          "num_coauthors", "num_publications", "num_citations",
                          "subjects", "affiliation_country", "affiliation_id",
                          "affiliation_name", "affiliation_type", "language",
                          "num_cited_refs"]
        if fields:
            invalid = [x for x in fields if x not in allowed_fields]
            if invalid:
                text = "Parameter fields contains invalid keywords: " +\
                       ", ".join(invalid)
                raise ValueError(text)
        else:
            fields = allowed_fields

        n_matches = len(self._matches)
        text = f"Providing information for {n_matches:,} " \
               f"match{get_ending(n_matches, 'es')}..."
        custom_print(text, verbose)
        matches = inform_matches(self, fields, verbose, refresh)
        self._matches = matches



"""Tests for establishing.fields_sources module."""

import pandas as pd

from sosia.establishing.constants import FIELD_SOURCE_MAP, SOURCE_INFO
from sosia.establishing.fields_sources import get_field_source_information

FIELD_SOURCE_MAP.unlink(missing_ok=True)
FIELD_SOURCE_MAP.unlink(missing_ok=True)
get_field_source_information()


def test_fields_sources_list():
    df = pd.read_csv(FIELD_SOURCE_MAP)
    assert isinstance(df, pd.DataFrame)
    assert list(df.columns) == ["source_id", "asjc"]
    assert df.shape[0] > 200_000


def test_sources_names_list():
    df = pd.read_csv(SOURCE_INFO, index_col=0)
    assert isinstance(df, pd.DataFrame)
    assert list(df.columns) == ["type", "title"]
    assert df.shape[0] > 90_000
    assert df.loc[12005, "type"] == "jr"
    assert df.loc[12005, "title"] == "Journal of Traumatic Stress"
    assert df.loc[90641, "type"] == "cp"
    expected_title = "Proceedings of the Conference on Traffic and Transportation Studies, ICTTS"
    assert df.loc[90641, "title"] == expected_title



"""Module that contains functions for retrieving data from a SQLite3 database cache."""

from sqlite3 import Connection
from typing import Optional, Union

import pandas as pd

from sosia.establishing import DB_TABLES


def drop_values(data: pd.DataFrame, conn: Connection, table: str) -> None:
    """Drop values from a database `table`."""
    fields = DB_TABLES[table]["primary"]
    if table == "author_data":
        fields = ("auth_id",)
    cursor = conn.cursor()
    if len(fields) == 1:
        values = data[fields[0]]
        q = f"DELETE FROM {table} WHERE {fields[0]} IN ({','.join(['?'] * len(values))})"
        cursor.execute(q, tuple(values))
    else:
        q = f"DELETE FROM {table} WHERE " + "=? AND ".join(fields) + "=?"
        cursor.executemany(q, data.to_records(index=False))
    conn.commit()


def insert_data(
    data: pd.DataFrame,
    conn: Connection,
    table: str,
) -> None:
    """Insert new information in SQL database.

    Parameters
    ----------
    data : DataFrame or 3-tuple
        Dataframe with author information.

    conn : sqlite3 connection
        Standing connection to a SQLite3 database.

    table : string
        The database table to insert into.  The query will be adjusted
        accordingly.

    Raises
    ------
    ValueError
        If parameter table is not one of the allowed values.
    """
    # Checks
    if table not in DB_TABLES:
        msg = f"table parameter must be one of {', '.join(DB_TABLES.keys())}"
        raise ValueError(msg)

    # Build query
    cols, _ = zip(*DB_TABLES[table]["columns"])
    values = ["?"] * len(cols)
    q = f"INSERT OR REPLACE INTO {table} ({','.join(cols)}) VALUES ({','.join(values)})"

    # Eventually tweak data
    if table in ('author_info', 'sources'):
        if data.empty:
            return None
    data = data[list(cols)]

    # Execute queries
    cursor = conn.cursor()
    cursor.executemany(q, data.to_records(index=False))
    conn.commit()


def retrieve_from_author_table(
        df: pd.DataFrame,
        conn: Connection,
        table: str,
        refresh: Union[bool, int] = False
) -> tuple[pd.DataFrame, list]:
    """Retrieve data on authors from specific `table` in SQL cache.

    Parameters
    ----------
    df : DataFrame
        DataFrame of authors to search.

    conn : sqlite3 connection
        Standing connection to a SQLite3 database.

    table : str
        The table of the SQLite3 database on which to perform the merge.

    refresh : bool, int (optional, default=False)
        All values other than False will lead to dropping database
        entries matching `df`.

    Returns
    -------
    incache : DataFrame
        Results found in cache.

    tosearch: list
        Results not found in cache.
    """
    # Set columns
    if table == "author_citations":
        cols = ["auth_id", "year"]
    else:
        cols = ["auth_id"]

    # Drop values if to be refreshed
    if not isinstance(refresh, bool) or refresh:
        drop_values(df, conn, table=table)

    # Query data
    incache = merge_query(df, conn, table=table, merge_cols=cols, join="INNER")
    tosearch = set(df["auth_id"].unique()) - set(incache["auth_id"].unique())
    return incache, sorted(tosearch)


def retrieve_authors_from_sourceyear(
        tosearch: pd.DataFrame,
        conn: Connection,
        drop: Optional[bool] = False
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """Search through sources by year for authors in SQL database.

    Parameters
    ----------
    tosearch : DataFrame
        DataFrame of source-year-combinations to be searched for.

    conn : sqlite3 connection
        Standing connection to a SQLite3 database.

    drop : bool, int (optional, default=False)
        Whether to drop matching information and replace them with data
        on disk.

    Returns
    -------
    data : DataFrame
        DataFrame in format ("source_id", "year", "auids"), where
        auids is a set of author IDs.

    missing: DataFrame
        DataFrame of source-year-combinations not in SQL database.
    """
    # Drop values if to be refreshed
    if drop:
        drop_values(tosearch, conn, table="sources")

    # Query authors for relevant journal-years
    cols = ["source_id", "year"]
    data = merge_query(tosearch.copy(), conn, table="sources",
                       merge_cols=cols, join="LEFT")

    # Finalize
    mask_missing = data["auids"].isna()
    incache = data[~mask_missing].reset_index(drop=True)
    missing = data.loc[mask_missing, cols].drop_duplicates().reset_index(drop=True)
    return incache, missing


def merge_query(
        df: pd.DataFrame,
        conn: Connection,
        table: str,
        merge_cols: list[str],
        join: str = "INNER"
) -> pd.DataFrame:
    """Query data from `table` matching `df` on `merge_cols`."""
    # Insert "temp" table
    df = df.astype({c: "int64" for c in merge_cols})
    cursor = conn.cursor()
    cursor.execute("DROP TABLE IF EXISTS temp")
    names = ", ".join(merge_cols)
    cursor.execute(f"CREATE TABLE temp ({names}, PRIMARY KEY({names}))")
    wildcards = ", ".join(["?"] * len(merge_cols))
    cursor.executemany(f"INSERT OR IGNORE INTO temp ({names}) VALUES ({wildcards})",
                       df.to_records(index=False))
    conn.commit()
    # Build query
    table_columns = [col[0] for col in DB_TABLES[table]["columns"]]
    b_columns = [col for col in table_columns if col not in merge_cols]
    a_select = ", ".join([f"a.{col}" for col in merge_cols])
    b_select = ", ".join([f"b.{col}" for col in b_columns])
    select_statement = f"{a_select}, {b_select}" if b_select else a_select
    conditions = " and ".join([f"a.{col} = b.{col}" for col in merge_cols])
    query = (
        f"SELECT {select_statement} FROM temp AS a "
        f"{join} JOIN {table} AS b "
        f"ON {conditions};"
    )
    # Retrieve data
    return pd.read_sql_query(query, conn)



"""Module with super class to represent a `Scientist`."""

from pathlib import Path
from typing import Optional, Union
from warnings import warn

from typing_extensions import Self

from pybliometrics.scopus import AbstractRetrieval, AffiliationRetrieval
from pybliometrics.scopus.exception import Scopus404Error

from sosia.establishing import connect_database, DEFAULT_DATABASE
from sosia.processing import add_source_names, base_query, count_citations, \
    extract_authors, find_main_affiliation, get_author_info, determine_main_field, \
    read_fields_sources_list
from sosia.utils import accepts


class Scientist(object):
    """Class to represent a scientist.
    """
    @property
    def affiliation_country(self) -> Optional[str]:
        """The country of the scientist's affiliation."""
        return self._affiliation_country

    @affiliation_country.setter
    @accepts(str)
    def affiliation_country(self, val: str) -> None:
        self._affiliation_country = val

    @property
    def affiliation_id(self) -> Optional[str]:
        """The affiliation ID (as string) of the scientist's most frequent
        affiliation in the active year.
        """
        return self._affiliation_id

    @affiliation_id.setter
    @accepts(str)
    def affiliation_id(self, val: str) -> None:
        self._affiliation_id = val

    @property
    def affiliation_name(self) -> Optional[str]:
        """The name of the scientist's affiliation."""
        return self._affiliation_name

    @affiliation_name.setter
    @accepts(str)
    def affiliation_name(self, val: str) -> None:
        self._affiliation_name = val

    @property
    def affiliation_type(self) -> Optional[str]:
        """The type of the scientist's affiliation."""
        return self._affiliation_type

    @affiliation_type.setter
    @accepts(str)
    def affiliation_type(self, val: str) -> None:
        self.affiliation_type = val

    @property
    def citations(self) -> Optional[int]:
        """The citation count of the scientist until the provided year."""
        return self._citations

    @citations.setter
    @accepts(int)
    def citations(self, val: int) -> None:
        self._citations = val

    @property
    def coauthors(self) -> list:
        """Sorted list of coauthors of the scientist on all publications
        until the comparison year.
        """
        return self._coauthors

    @coauthors.setter
    @accepts((set, list, tuple))
    def coauthors(self, val: list) -> None:
        self._coauthors = sorted(val)

    @property
    def fields(self) -> Union[set, list, tuple]:
        """The fields of the scientist until the provided year, estimated from
        the sources (journals, books, etc.) she published in.
        """
        return self._fields

    @fields.setter
    @accepts((set, list, tuple))
    def fields(self, val: Union[set, list, tuple]) -> None:
        self._fields = val

    @property
    def first_year(self) -> int:
        """The scientist's year of first publication."""
        return self._first_year

    @first_year.setter
    @accepts(int)
    def first_year(self, val: int) -> None:
        self._first_year = val

    @property
    def first_name(self) -> Optional[str]:
        """The scientist's first name."""
        return self._first_name

    @first_name.setter
    @accepts(str)
    def first_name(self, val: str) -> None:
        self._name = val

    @property
    def main_field(self) -> tuple:
        """The scientist's main field of research, as tuple in
        the form (ASJC code, general category).

        The main field is the field with the most publications, provided it
        is not Multidisciplinary (ASJC code 1000).  In case of an equal number
        of publications, preference is given to non-general fields (those
        whose ASJC ends on a digit other than 0).
        """
        return self._main_field

    @main_field.setter
    def main_field(self, val: tuple) -> None:
        if not isinstance(val, tuple) or len(val) != 2:
            raise TypeError("Value must be a two-element tuple.")
        self._main_field = val

    @property
    def name(self) -> Optional[str]:
        """The scientist's complete name."""
        return self._name

    @name.setter
    @accepts(str)
    def name(self, val: str) -> None:
        self._name = val

    @property
    def language(self) -> Optional[str]:
        """The language(s) the scientist published in."""
        return self._language

    @language.setter
    @accepts(str)
    def language(self, val: str) -> None:
        self._language = val

    @property
    def last_year(self) -> int:
        """The scientist's most recent year with publication(s) before the
         match year (which may be the same).
         """
        return self._last_year

    @last_year.setter
    @accepts(int)
    def last_year(self, val: int) -> None:
        self._last_year = val

    @property
    def publications(self) -> Union[set, list, tuple]:
        """List of the scientists' publications."""
        return self._publications

    @publications.setter
    @accepts((set, list, tuple))
    def publications(self, val: Union[set, list, tuple]) -> None:
        self._publications = val

    @property
    def sources(self) -> Union[list, tuple]:
        """The Scopus IDs of sources (journals, books, etc.) in which the
        scientist published in.
        """
        return self._sources

    @sources.setter
    @accepts((list, tuple))
    def sources(self, val: Union[list, tuple]) -> None:
        self._sources = add_source_names(val, self.source_names)

    @property
    def surname(self) -> Optional[str]:
        """The scientist's surname."""
        return self._surname

    @surname.setter
    @accepts(str)
    def surname(self, val: str):
        self._name = val

    @property
    def subjects(self) -> Union[set, list, tuple]:
        """The subject areas of the scientist's publications."""
        return self._subjects

    @subjects.setter
    @accepts((set, list, tuple))
    def subjects(self, val: Union[set, list, tuple]) -> None:
        self._subjects = val

    def __init__(
        self,
        identifier: list[int],
        year: Union[str, int],
        refresh: Union[bool, int] = False,
        eids: Optional[list[str]] = None,
        db_path: Optional[Union[str, Path]] = None,
        verbose: Optional[bool] = False
    ) -> None:
        """Class to represent a scientist.

        Parameters
        ----------
        identifier : list of int
            List of Scopus Author IDs of the scientist.

        year : str or numeric
            Year for which characteristics should be defined for.

        refresh : boolean or int (optional, default=False)
            Whether to refresh cached results (if they exist) or not. If int
            is passed, results will be refreshed if they are older than
            that value in number of days.

        eids : list (optional, default=None)
            A list of Scopus EIDs of the publications of the scientist.  If
            provided, the scientist's properties are inferred from these
            publications, instead of the list of publications obtained from
            the Scopus Author ID(s).

        db_path : str or pathlib.Path (optional, default=None)
            The path of the local SQLite database to connect to.  If None,
            will default to `~/.cache/sosia/main.sqlite`.  Will be created
            if the database doesn't exist.

        verbose : bool (optional, default=False)
            Whether to report on the initialization process.

        Raises
        ------
        Exception
            When there are no publications for the author until the
            provided year.
        """
        self.identifier = identifier
        self.year = int(year)

        # Local database (cache)
        if not db_path:
            db_path = DEFAULT_DATABASE
        db_path = Path(db_path)
        self.sql_conn = connect_database(db_path, verbose=verbose)

        # Read mapping of fields to sources
        fields, info = read_fields_sources_list(verbose=verbose)
        self.field_source = fields
        self.source_info = info
        source_names = self.source_info.set_index("source_id")["title"].to_dict()
        self.source_names = source_names

        # Load list of publications
        if eids:
            q = f"EID({' OR '.join(eids)})"
        else:
            q = f"AU-ID({') OR AU-ID('.join([str(i) for i in identifier])})"
        integrity_fields = ["eid", "author_ids", "coverDate", "source_id"]
        res = base_query("docs", q, refresh, fields=integrity_fields)
        self._publications = [p for p in res if int(p.coverDate[:4]) <= int(year)]
        if not self._publications:
            text = "No publications found for author "\
                   f"{'-'.join([str(i) for i in identifier])} until {self.year}"
            raise ValueError(text)
        self._eids = eids or [p.eid for p in self._publications]

        # Publication range
        pub_years = [int(p.coverDate[:4]) for p in self._publications if p.coverDate]
        self._first_year = min(pub_years)
        self._last_year = max(pub_years)

        # Count of citations
        search_ids = eids or identifier
        self._citations = count_citations(search_ids, self.year+1, identifier)

        # Coauthors
        self._coauthors = sorted(set(extract_authors(self._publications)) - set(identifier))

        # Author search information
        source_ids = set([int(p.source_id) for p in self._publications
                          if p.source_id])
        self._sources = add_source_names(sorted(source_ids), self.source_names)
        mask = fields["source_id"].isin(source_ids)
        self._fields = fields[mask]["asjc"].astype(int).tolist()
        self._main_field = determine_main_field(self._fields)
        if not self._main_field[0]:
            text = "Not possible to determine research field(s) of "\
                   "researcher.  Functionality is reduced."
            warn(text, UserWarning)

        # Most recent geolocation
        afid = find_main_affiliation(identifier, self._publications, year)
        self._affiliation_id = afid
        try:
            aff = AffiliationRetrieval(afid, refresh=refresh)
            self._affiliation_country = aff.country
            self._affiliation_name = aff.affiliation_name
            self._affiliation_type = aff.org_type
        except (Scopus404Error, ValueError):
            self._affiliation_country = None
            self._affiliation_name = None
            self._affiliation_type = None
        self._language = None

        # Author name from profile with most documents
        df = get_author_info(self.identifier, self.sql_conn,
                             refresh=refresh, verbose=False)
        au = df.sort_values("documents", ascending=False).iloc[0]
        self._subjects = [a.split(" ")[0] for a in au.areas.split("; ")]
        self._surname = au.surname or None
        self._first_name = au.givenname or None
        name = ", ".join([self._surname or "", au.givenname or ""])
        if name == ", ":
            name = None
        self._name = name

    def get_publication_languages(self, refresh: bool = False) -> Self:
        """Parse languages of published documents."""
        langs = set()
        for eid in self._eids:
            try:
                ab = AbstractRetrieval(eid, view="FULL", refresh=refresh)
            except Scopus404Error:
                continue
            langs.add(ab.language)
        self._language = "; ".join(sorted(filter(None, langs)))
        return self



"""Module with decorator to check types of property."""


def accepts(*classinfo_args):
    """Decorator to check types of property."""
    def isinstance_decorator_wrapper(old_fn):
        def new_fn(self, *args, **kwargs):
            for i, classinfo in enumerate(classinfo_args):
                arg = args[i]
                if not isinstance(arg, classinfo):
                    if isinstance(classinfo, tuple):
                        obj_type = "' or '".join([x.__name__ for x in classinfo])
                    else:
                        obj_type = classinfo.__name__
                    msg = f"Attribute {old_fn.__name__} must be of type "\
                          f"'{obj_type}' but '{type(arg).__name__}' was passed"
                    raise TypeError(msg)
            return old_fn(self, *args, **kwargs)
        return new_fn
    return isinstance_decorator_wrapper



"""Shared pytest fixtures and configurations for the test suite."""

import pytest
from pathlib import Path

from pybliometrics.scopus import init
from sosia.classes import Original, Scientist
from sosia.establishing import connect_database, make_database

init()


@pytest.fixture(scope="session")
def test_cache():
    test_cache = Path.home() / ".cache" / "sosia" / "test.sqlite"
    test_cache.unlink(missing_ok=True)
    make_database(test_cache)
    return test_cache


@pytest.fixture
def test_conn(test_cache):
    conn = connect_database(test_cache, verbose=False)
    yield conn
    conn.close()


@pytest.fixture(scope="session")
def refresh_interval():
    return 30


@pytest.fixture(scope="session")
def scientist1(test_cache, refresh_interval):
    return Scientist([6701809842], 2001, refresh=refresh_interval,
                     db_path=test_cache)


@pytest.fixture(scope="session")
def scientist2(test_cache, refresh_interval):
    return Scientist([55208373700, 55208373700], 2017,
                     refresh=refresh_interval, db_path=test_cache)


@pytest.fixture(scope="session")
def scientist3(test_cache, refresh_interval):
    eids = ["2-s2.0-84959420483", "2-s2.0-84949113230"]
    return Scientist([55208373700], 2017, eids=eids,
                     refresh=refresh_interval, db_path=test_cache)


@pytest.fixture(scope="session")
def scientist4(test_cache, refresh_interval):
    return Scientist([55208373700], 2015, refresh=refresh_interval,
                     db_path=test_cache)


@pytest.fixture(scope="session")
def original1(test_cache, refresh_interval):
    return Original(55208373700, 2018, db_path=test_cache,
                    refresh=refresh_interval)



"""Module with functions for extracting information from publications and matching scientists."""

from collections import defaultdict, Counter, namedtuple
from typing import Optional

import pandas as pd
from pybliometrics.scopus import AbstractRetrieval
from pybliometrics.scopus.exception import Scopus404Error
from tqdm import tqdm

from sosia.processing.constants import ASJC_2D
from sosia.processing.utils import compute_overlap
from sosia.processing.querying import base_query


def extract_authors(pubs):
    """Get list of author IDs from a list of namedtuples representing
    publications.
    """
    auths = [x.author_ids.split(";") for x in pubs if isinstance(x.author_ids, str)]
    return [int(au) for sl in auths for au in sl]


def extract_yearly_author_data(auth_id: int, *args, **kwargs) -> pd.DataFrame:
    """Extract relevant information from an author's yearly publication list.

    Parameters
    ----------
    auth_id : int
        The Scopus Author ID.

    *args, **kwargs : tuple or dict (optional)
        Additional options passed on to `base_query()`: `refresh`.

    Returns
    -------
    out : DataFrame
        DataFrame with yearly information (first year of publication,
        publication stock, and coauthor stock) on requested author.
    """
    # Get data
    fields = ["eid", "creator", "author_ids"]
    docs = base_query("docs", f"AU-ID({auth_id})", fields=fields,
                      *args, **kwargs)
    cols = ["eid", "coverDate", "author_ids"]
    df = pd.DataFrame(docs)[cols].rename(columns={"coverDate": "year"})
    df["year"] = df["year"].str[:4].astype("uint16")
    # First year
    first_year = df["year"].min()
    # Publications
    pub_counts = df['year'].value_counts()
    # Coauthors
    authors = (df.set_index(["eid", "year"])
                 ['author_ids'].str.split(';', expand=True)
                 .melt(ignore_index=False, value_name='author_id')
                 .dropna()
                 .reset_index().drop(columns=['variable', "eid"])
                 .sort_values("year", ascending=True)
                 .drop_duplicates())
    unique_authors = set()
    coauth_counts = defaultdict(int)
    for year, subset in authors.groupby('year'):
        subset = subset[subset["author_id"] != str(auth_id)]
        unique_authors.update(subset['author_id'].unique())
        coauth_counts[year] = len(unique_authors)
    # Combine
    data = {"auth_id": auth_id, "year": range(first_year, df["year"].max() + 1)}
    out = pd.DataFrame(data)
    out.index = out["year"]
    out["first_year"] = int(first_year)
    out["n_pubs"] = pub_counts
    out["n_pubs"] = out["n_pubs"].fillna(0).cumsum().astype(int)
    out["n_coauth"] = coauth_counts
    out["n_coauth"] = out["n_coauth"].ffill().astype(int)
    return out.reset_index(drop=True)


def find_main_affiliation(auth_ids, pubs, year):
    """Find the most common affiliation ID of a scientist on publications
    with valid information of the most recent year.

    Parameters
    ----------
    auth_ids : list of str
        A list of Scopus Author Profile IDs for which the affiliation should
        be searched for.

    pubs : list of namedtuple
        The publications associated with the Author IDs as returned from a
        scopus query.

    year : int
        The year for which we would like to have the country.

    Returns
    -------
    affiliation_id : str or None
        The most common affiliation_id of the scientist in the year closest
        to the provided year, given that the publications list valid
        information for each output.  Equals None when no valid
        publications are found.
    """
    pubs = [p for p in pubs if p.author_ids and p.author_afids]
    # Find affiliation ID of all available publications
    affs = defaultdict(lambda: Counter())
    for p in pubs:
        cur_year = int(p.coverDate[:4])
        if cur_year > year:
            continue
        authors = [int(a) for a in p.author_ids.split(";")]
        for focal in set(auth_ids).intersection(authors):
            idx = authors.index(focal)
        try:
            aff_ids = p.author_afids.split(";")[idx].split("-")
        except (IndexError, UnboundLocalError):
            continue
        affs[cur_year].update(Counter([a for a in aff_ids if a]))
    # Use only most recent publications
    try:
        max_year = max(affs.keys())
        main_aff = affs[max_year].most_common()[0][0]
    except (IndexError, ValueError):
        main_aff = None
    return main_aff


def determine_main_field(fields) -> tuple[Optional[int], Optional[str]]:
    """Determine the most common field (4-digit ASJC) and most common discipline.

    Parameters
    ----------
    fields : iterable of int
        Lists of fields the researcher is active in.

    Returns
    -------
    main : int
        The most common 4-digit ASJC field.

    name : str
        The discipline of the main field.

    Note
    ----
    We exclude multidisciplinary and give preference to non-general fields.
    """
    # Exclude Multidisciplinary
    while 1000 in fields:
        fields.remove(1000)

    # Verify at least some information is present
    if not fields:
        return None, None

    # 4 digit field
    c = Counter(fields)
    top_fields = [f for f, val in c.items() if val == max(c.values())]
    if len(top_fields) == 1:
        main_4 = top_fields[0]
    else:
        non_general_fields = [f for f in top_fields if f % 1000 != 0]
        if non_general_fields:
            main_4 = non_general_fields[0]
        else:
            main_4 = top_fields[0]

    # 2 digit field
    c = Counter([str(f)[:2] for f in fields])
    main_2 = int(c.most_common(1)[0][0])
    name = ASJC_2D[main_2]

    return main_4, name


def inform_match(profile, keywords, refresh):
    """Create namedtuple adding information to matches.

    Parameters
    ----------
    profile : sosia.Scientist()
        A Scientist() object representing a match.

    keywords : iterable of strings
        Which information to add to the match.

    refresh : bool
        Whether to refresh all cached files or not.

    Returns
    -------
    match_info : dict
        Information corresponding to provided keywords.
    """
    info = {
        "ID": profile.identifier[0],
        "name": profile.name,
        "first_name": profile.first_name,
        "surname": profile.surname,
        "first_year": profile.first_year,
        "last_year": profile.last_year,
        "num_coauthors": len(profile.coauthors),
        "num_publications": len(profile.publications),
        "num_citations": profile.citations,
        "subjects": profile.subjects,
        "affiliation_country": profile.affiliation_country,
        "affiliation_id": profile.affiliation_id,
        "affiliation_name": profile.affiliation_name,
        "affiliation_type": profile.affiliation_type,
    }
    match_info = {k: v for k, v in info.items() if k in keywords + ["ID", "name"]}
    if "language" in keywords:
        lang = profile.get_publication_languages(refresh=refresh).language
        match_info["language"] = lang
    return match_info


def inform_matches(self, keywords, verbose, refresh):
    """Add match-specific information to all matches.

    Parameters
    ----------
    self : sosia.Original()
        Object whose matches should receive additional information.

    keywords : iterable of strings
        Which information to add to matches.

    verbose : bool
        Whether to report on the progress of the process and the completeness
        of document information.

    refresh : bool
        Whether to refresh cached results (if they exist) or not. If
            int is passed, results will be refreshed if they are older
            than that value in number of days.

    Returns
    -------
    out : list of namedtuples
        A list of namedtuples representing matches.  Provided information
        depend on provided keywords.
    """
    from sosia.classes import Scientist

    # Create Match object
    fields = "ID name " + " ".join(keywords)
    m = namedtuple("Match", fields)

    # Preparation
    doc_parse = "num_cited_refs" in keywords
    if doc_parse:
        focal_docs = parse_docs([d.eid for d in self.publications], refresh)
        focal_refs, focal_refs_n = focal_docs

    # Add selected information match-by-match
    out = []
    completeness = {}
    for auth_id in tqdm(self.matches, disable=not verbose):
        p = Scientist([auth_id], self.year, refresh=refresh,
                      db_path=self.sql_fname)
        match_info = inform_match(p, keywords, refresh=refresh)
        # Abstract and reference similarity is performed jointly
        if doc_parse:
            eids = [d.eid for d in p.publications]
            refs, refs_n = parse_docs(eids, refresh=refresh)
            completeness[auth_id] = (refs_n, len(eids))
            if "num_cited_refs" in keywords:
                ref_cos = compute_overlap(refs, focal_refs)
                match_info["num_cited_refs"] = ref_cos
        out.append(m(**match_info))

    # Eventually print information on missing information
    if verbose and doc_parse:
        for auth_id, completeness in completeness.items():
            _print_missing_docs([auth_id], completeness[0], completeness[1])
        focal_pubs_n = len(self.publications)
        _print_missing_docs(self.identifier, focal_refs_n, focal_pubs_n,
                            res_type="Original")
    return out


def parse_docs(eids, *args, **kwargs):
    """Find the set of references of provided articles.

    Parameters
    ----------
    eids : list of str
        Scopus Document EIDs representing documents to be considered.

    *args, **kwargs : tuple or dict (optional)
        Additional options passed on to `AbstractRetrieval()`: `refresh`.

    Returns
    -------
    refs : set
        The set of Scopus Document EIDs of cited references.

    n_valid_refs : int
        The number of documents with valid reference information.
    """
    docs = []
    for eid in eids:
        try:
            docs.append(AbstractRetrieval(eid, view="FULL", *args, **kwargs))
        except Scopus404Error:
            continue
    ref_lst = [ab.references for ab in docs if ab.references]
    valid_refs = len(ref_lst)
    ref_ids = [ref.id for sl in ref_lst for ref in sl]
    refs = set(filter(None, ref_ids))
    return refs, valid_refs


def _print_missing_docs(auth_id, n_valid_refs, total, res_type="Match"):
    """Auxiliary function to print information on reference lists."""
    auth_ids = [str(a) for a in auth_id]
    prefix = f"{res_type} {';'.join(auth_ids)}:"
    missing = total-n_valid_refs
    if missing:
        plural = "s" if missing > 1 else ""
        middle = f"{missing} reference list{plural} out"
    else:
        middle = "No reference list"
    suffix = f"of {total} documents missing"
    text = " ".join([prefix, middle, suffix])
    print(text)



"""This module provides functions for connecting to and creating a SQLite database."""

import sqlite3
from pathlib import Path
from typing import Optional

from numpy import int32, int64

from sosia.utils import custom_print


def connect_database(fname: Path, verbose) -> sqlite3.Connection:
    """Connect to local SQLite3 database to be used as cache.

    Parameters
    ----------
    fname : pathlib.Path
        The path of the SQLite3 database to connect to.

    verbose : bool (optional, default=False)
        Whether to report on the status of the database.
    """
    for val in (int32, int64):
        sqlite3.register_adapter(val, int)

    if not fname.exists():
        make_database(fname, verbose=verbose)
    text = f"Connection to local database '{fname}' established"
    custom_print(text, verbose)

    return sqlite3.connect(fname)


def make_database(
        fname: Optional[Path] = None,
        verbose: bool = False,
        drop: bool = False
) -> None:
    """Make SQLite database with predefined tables and keys.

    Parameters
    ----------
    fname : pathlib.Path (optional, default=None)
        The path of the SQLite database to connect to.  If None, will default
        to `~/.cache/sosia/main.sqlite`.

    verbose : boolean (optional, default=False)
        Whether to report on the progess of the process.

    drop : boolean (optional, default=False)
        If True, deletes and recreates all tables in cache (irreversible).
    """
    from sosia.establishing.constants import DB_TABLES, DEFAULT_DATABASE

    if not fname:
        fname = DEFAULT_DATABASE

    # Create database
    existed = fname.exists()
    fname.parent.mkdir(parents=True, exist_ok=True)
    conn = sqlite3.connect(fname)

    # Create tables
    cursor = conn.cursor()
    for table, variables in DB_TABLES.items():
        if drop:
            cursor.execute(f"DROP TABLE IF EXISTS {table}")
        columns = ", ".join(" ".join(v) for v in variables["columns"])
        q = f"CREATE TABLE IF NOT EXISTS {table} "\
            f"({columns}, PRIMARY KEY({', '.join(variables['primary'])}))"
        cursor.execute(q)

    # Report progress
    if fname.exists():
        if existed:
            if drop:
                msg = f"Local database '{fname}' re-created successfully"
            else:
                msg = f"Local database '{fname}' not changed"
        else:
            msg = f"Local database '{fname}' created successfully"
    else:
        msg = f"Failed to create the local database '{fname}'"
    custom_print(msg, verbose)



"""Module with utility functions for processing data in sosia."""

from itertools import islice
from math import ceil
from typing import Union


def chunk_list(data: list, size: int) -> list[list]:
    """Chunk a list into bins of a given size and merge the last if necessary."""
    data_iter = iter(data)
    chunks = list(iter(lambda: list(islice(data_iter, size)), []))

    # Merge last chunk into previous if smaller than half the chunk size
    if len(chunks[-1]) <= size // 2 and len(chunks) > 1:
        chunks[-2].extend(chunks.pop())

    return chunks


def compute_overlap(left, right):
    """Compute overlap of two sets in a robust way."""
    if not left and not right:
        return None
    return len(left.intersection(right))


def flat_set_from_df(df, col):
    """Flatten Series from DataFrame which contains lists and
    return as set, optionally after filtering the DataFrame.
    """
    lists = df[col].tolist()
    return set([item for sublist in lists for item in sublist])


def generate_filter_message(number: int, margins: tuple, label: str):
    """Generate filter progress message."""
    if number == 1:
        plural = ""
    else:
        plural = "s"
    if margins[0] == margins[1]:
        qualifier = "same"
        if "year" in label:
            suffix = margins[0]
        else:
            suffix = f"{margins[0]:,}"
    else:
        qualifier = "similar"
        if "year" in label:
            suffix = f"{min(margins)} to {max(margins)}"
        else:
            suffix = f"{min(margins):,} to {max(margins):,}"
    text = (f"... left with {number:,} candidate{plural} with "
            f"{qualifier} {label} ({suffix})")
    return text


def compute_margins(base: int, val: Union[float, int]) -> tuple[int, int]:
    """Computer upper and lower margin around a base value.

    Parameters
    ----------
    base : int
        The value around which a margin should be created.

    val : int or float
        The margin size.  If float, val will be interpreted as percentage.

    Returns
    -------
    min, may : tuple
        A tuple with the upper and lower value.
    """
    if isinstance(val, float):
        margin = ceil(val * base)
        t = (max(base - margin, 0), base + margin)
    elif isinstance(val, int):
        t = (max(base - val, 0), base + val)
    else:
        raise TypeError("Value must be either float or int.")
    return t



"""Tests for processing.getting module."""

from pandas import DataFrame

from sosia.processing.getting import get_author_data, get_author_info, \
    get_citations


def test_get_author_data(test_conn):
    group = [6701809842, 16319073600, 54984906100, 56148489300, 57131011400,
             57194816659, 35097480000, 56055501900, 20434039300, 6602070937]
    data = get_author_data(group, test_conn)
    assert isinstance(data, DataFrame)
    expected_cols = ['auth_id', 'year', 'first_year', 'n_pubs', 'n_coauth']
    assert list(data.columns) == expected_cols
    assert data.shape[0] > 200
    assert data["auth_id"].nunique() == len(group)


def test_get_author_info(test_conn, refresh_interval):
    auth_list = [6701809842, 55208373700]
    auth_data = get_author_info(auth_list, test_conn, refresh=refresh_interval)
    assert isinstance(auth_data,  DataFrame)
    expected_cols = ["auth_id", "surname", "givenname", "documents", "areas"]
    assert auth_data.columns.tolist() == expected_cols
    assert auth_data["auth_id"].tolist() == auth_list
    assert auth_data["surname"].tolist() == ["Harhoff", "Baruffaldi"]


def test_get_citations(test_conn):
    group = [6701809842, 16319073600, 6602070937]
    data = get_citations(group, year=2023, conn=test_conn)
    assert isinstance(data, DataFrame)
    expected_cols = ['auth_id', 'n_cits']
    assert list(data.columns) == expected_cols
    assert data["auth_id"].nunique() == len(group)
    assert data.shape[0] == len(group)
    assert data.loc[0, "n_cits"] > 4000



"""Module with functions related to initializing the sosia processing."""

from typing import Optional, Union

from pandas import read_csv

from sosia.establishing.constants import FIELD_SOURCE_MAP, SOURCE_INFO
from sosia.establishing.fields_sources import get_field_source_information
from sosia.utils import custom_print


def add_source_names(
        source_ids: list[Union[str, int, tuple]],
        names: dict
) -> list[tuple[Union[str, int, tuple], Optional[str]]]:
    """Add names of sources to list of source IDs turning the list into a
    list of tuples.
    """
    if isinstance(source_ids[0], tuple):
        return source_ids
    else:
        return [(s_id, names.get(int(s_id))) for s_id in sorted(set(source_ids))]


def read_fields_sources_list(verbose: bool = False):
    """Auxiliary function to read FIELD_SOURCE_MAP and create it before,
    if necessary.
    """
    try:
        field = read_csv(FIELD_SOURCE_MAP)
        info = read_csv(SOURCE_INFO)
        text = f"Using information for {info.shape[0]:,} sources as well as "\
               f"{field.shape[0]:,} field-source assignments from '{SOURCE_INFO.parent}'"
        custom_print(text, verbose)
    except FileNotFoundError:
        get_field_source_information(verbose=verbose)
        field = read_csv(FIELD_SOURCE_MAP)
        info = read_csv(SOURCE_INFO)
    return field, info



"""Module to configure the logging for the package."""

import logging
from typing import Union
from pathlib import Path

from pybliometrics.scopus import AuthorSearch, ScopusSearch

logger = None


def create_logger(log_file: Union[str, Path]) -> None:
    """Configure a logger."""
    global logger

    logger = logging.getLogger("sosia")

    log_file.parent.mkdir(parents=True, exist_ok=True)
    file_handler = logging.FileHandler(log_file, mode='w', encoding='utf-8')
    formatter = logging.Formatter(
        '%(asctime)s - %(levelname)s: %(message)s',
        datefmt='%y-%m-%d %H:%M:%S'
    )
    file_handler.setFormatter(formatter)

    logger.addHandler(file_handler)
    logger.setLevel(logging.DEBUG)
    logger.propagate = False


class ScopusLogger:
    """Context manager to log scopus"""
    def __init__(self, scopus_api, params):
        self.scopus_obj: Union[AuthorSearch, ScopusSearch]
        self.scopus_api = scopus_api
        self.query = params.get('query')
        self.view = params.get('view', 'Default')

        if logger is None:
            create_logger()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        if not exc_type:
            results = self.scopus_obj.get_results_size()
            self.view = self.scopus_obj._view
        logger.debug(
            '\n\t- Scopus API: %s with %s view \n\t- Query: %s\n\t- Results: %s',
            self.scopus_api, self.view, self.query, exc_type or results
        )



"""This file is the initialization module for the `sosia.classes`. It imports the `Scientist` and
`Original` classes from the respective modules.
"""

from sosia.classes.scientist import *
from sosia.classes.original import *



"""Tests for class `Scientist`."""


def test_affiliation_country(scientist1, scientist2, scientist3, scientist4):
    assert scientist1.affiliation_country == "Germany"
    assert scientist2.affiliation_country == "Switzerland"
    assert scientist3.affiliation_country == "Switzerland"
    assert scientist4.affiliation_country == "Switzerland"


def test_affiliation_id(scientist1, scientist2, scientist3, scientist4):
    assert scientist1.affiliation_id == '60028717'
    assert scientist2.affiliation_id == '60028186'
    assert scientist3.affiliation_id == '60028186'
    assert scientist4.affiliation_id == '60028186'


def test_affiliation_name(scientist1, scientist2, scientist3, scientist4):
    lmu = 'Ludwig-Maximilians-Universität München'
    assert scientist1.affiliation_name == lmu
    epfl = 'École Polytechnique Fédérale de Lausanne'
    assert scientist2.affiliation_name == epfl
    assert scientist3.affiliation_name == epfl
    assert scientist4.affiliation_name == epfl


def test_affiliation_type(scientist1, scientist2, scientist3, scientist4):
    assert scientist1.affiliation_type == 'univ'
    assert scientist2.affiliation_type == 'univ'
    assert scientist3.affiliation_type == 'univ'
    assert scientist4.affiliation_type == 'univ'


def test_citations(scientist1, scientist2, scientist3, scientist4):
    assert scientist1.citations >= 47
    assert scientist2.citations >= 28
    assert scientist3.citations >= 2
    assert scientist4.citations >= 19


def test_coauthors(scientist1, scientist2, scientist3):
    assert isinstance(scientist1.coauthors, list)
    expected = [6506426539, 6506571902, 6506756510, 6602701792, 6701494844,
                7004064836, 7005044638, 7101829476, 11042582400, 24364642400,
                35838036900]
    assert len(scientist1.coauthors) == len(expected)
    for coauth in expected:
        assert coauth in scientist1.coauthors
    assert isinstance(scientist2.coauthors, list)
    expected = [24464562500, 24781156100, 36617057700, 54929867200,
                54930777900, 55875219200, 57217825601]
    assert len(scientist2.coauthors) == len(expected)
    for coauth in expected:
        assert coauth in scientist2.coauthors
    assert isinstance(scientist3.coauthors, list)
    expected = [36617057700, 54929867200, 54930777900, 55875219200]
    assert len(scientist3.coauthors) == len(expected)
    for coauth in expected:
        assert coauth in scientist3.coauthors
    expected = [36617057700, 54929867200, 54930777900, 55875219200]
    assert len(scientist3.coauthors) == len(expected)


def test_fields(scientist1, scientist2, scientist3):
    expected = [2002, 2003, 2002, 2308, 1408, 1803, 1405, 1408, 1803, 2002,
                3301, 1400, 1402, 2002, 2002, 3317, 1400, 2002, 1405, 2000]
    assert set(scientist1.fields) == set(expected)
    expected = [2300, 3300, 2002, 1405, 1408, 1803, 1400, 1405]
    assert set(scientist2.fields) == set(expected)
    expected = [2002, 1405, 1408, 1803]
    assert set(scientist3.fields) == set(expected)


def test_first_year(scientist1, scientist2, scientist3):
    assert scientist1.first_year == 1996
    assert scientist2.first_year == 2012
    assert scientist3.first_year == 2016


def test_last_year(scientist1, scientist2, scientist3, scientist4):
    assert scientist1.last_year == 2001
    assert scientist2.last_year == 2017
    assert scientist3.last_year == 2016
    assert scientist4.last_year == 2012


def test_sources(scientist1, scientist2, scientist3):
    received = scientist1.sources
    expected = [(17472, 'Journal of Banking and Finance'),
                (20022, 'Economic Policy'),
                (21307, 'Management Science'),
                (22900, 'Research Policy'),
                (24204, 'Review of Economics and Statistics'),
                (24389, 'Journal of Industrial Economics'),
                (26878, 'Journal of Population Economics'),
                (28994, 'Journal of Evolutionary Economics'),
                (10600153365, 'Economics of Innovation and New Technology')]
    assert len(received) == len(expected)
    for e in expected:
        assert e in received
    received = scientist2.sources
    expected = [(23013, 'Industry and Innovation'),
                (21100858668, None),
                (15143, 'Regional Studies'),
                (18769, 'Applied Economics Letters'),
                (22900, 'Research Policy')]
    assert len(received) == len(expected)
    for e in expected:
        assert e in received
    received = scientist3.sources
    expected = [(18769, "Applied Economics Letters"),
                (22900, "Research Policy")]
    assert len(received) == len(expected)
    for e in expected:
        assert e in received


def test_sources_change(scientist1):
    backup = scientist1.sources
    expected = [(14351, "Brain Research Reviews"),
                (18632, "Progress in Brain Research")]
    scientist1.sources, _ = zip(*expected)
    assert scientist1.sources == expected
    scientist1.sources = backup


def test_main_field(scientist1, scientist2, scientist3):
    assert scientist1.main_field == (2002, "ECON")
    assert scientist2.main_field == (1405, "BUSI")
    assert scientist3.main_field == (2002, "BUSI")


def test_name(scientist1, scientist2, scientist3):
    assert scientist1.name == "Harhoff, Dietmar"
    assert scientist2.name == "Baruffaldi, Stefano Horst"
    assert scientist3.name == "Baruffaldi, Stefano Horst"


def test_publications(scientist1, scientist2, scientist3):
    cols = ["eid", "affiliation_country", "author_names", "author_ids",
            "author_afids", "coverDate", "authkeywords",
            "citedby_count", "description"]
    # scientist1
    received = scientist1.publications
    assert len(received) >= 11
    assert received[2].eid == '2-s2.0-0001093103'
    for p in received:
        assert all(c in p._fields for c in cols)
    # scientist2
    received = scientist2.publications
    assert len(received) >= 7
    assert received[-1].eid == "2-s2.0-84866317084"
    for p in received:
        assert all(c in p._fields for c in cols)
    # scientist3
    received = scientist3.publications
    assert len(received) >= 2
    received = scientist2.publications
    assert len(received) >= 7
    assert received[0].eid == "2-s2.0-85015636484"
    for p in received:
        assert all(c in p._fields for c in cols)


def test_publication_languages(scientist1, scientist2, scientist3):
    assert scientist1.language is None
    scientist1.get_publication_languages()
    assert scientist1.language == "eng"
    scientist3.get_publication_languages()
    assert scientist3.language == "eng"



"""This module imports various processing functionalities for Sosia."""

from sosia.processing.caching import *
from sosia.processing.constants import *
from sosia.processing.extracting import *
from sosia.processing.getting import *
from sosia.processing.initializing import *
from sosia.processing.querying import *
from sosia.processing.utils import *
